/*
 * Adapted for Windows 95/98/NT and Borland C++ 5.0
 * by Thomas Nystrm, Stacken Computer Club
 * 1998-12-30 20:47:46
 */	 

/*
 * Copyright (c) 1995-2000 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* $Id: AuthOption.cpp,v 1.4 1998/04/04 17:55:57 assar Exp $ */

//
//
// AuthOption
//
//

#include "common.h"
#pragma hdrstop

#include "authopt.h"
#include "tncodes.h"
#include "emul.rh"
#include <krb.h>

#include "kerb5.h"
#include "kclient.h"

extern Kerberos5Context *k5c;

static unsigned char str_data[1024] = { IAC, SB, AUTHENTICATION, IS,
			  		KERBEROS_V5, };
AuthOption::AuthOption(KTelnetConnection *c, ConnectionData _cd)
	: Option (c, AUTHENTICATION)
{
	cd = _cd;
	pKClient = c->pKClient;
	AuthMode = enAuthUnk;
	AuthIdx = AuthCount = 0;
}

AuthOption::~AuthOption()
{
}

bool
AuthOption::KerbInit()
{
	return (pKClient == 0);
}

int
AuthOption::Data(char mod, int type, void *d, int c)
{
	unsigned char *p = str_data + 4;
	unsigned char *cd = (unsigned char *)d;

	if (c == -1)
		c = strlen((char *)cd);

	*p++ = KERBEROS_V5;
	*p++ = mod;
	*p++ = (unsigned char)type;
	while (c-- > 0) {
		if ((*p++ = *cd++) == (unsigned char)IAC)
			*p++ = IAC;
		}
	*p++ = IAC;
	*p++ = SE;

	SendReply(str_data, p - str_data);
	return 0;
}

void
AuthOption::SendUserId()
{
	// Send name
	char nameBuff[4];
	char iacseBuff[2];
	nameBuff[0] = IAC;
	nameBuff[1] = SB;
	nameBuff[2] = AUTHENTICATION;
	nameBuff[3] = NAME;
	SendReply(nameBuff, 4);
	SendReply(cd.UserName, strlen(cd.UserName));
	iacseBuff[0] = IAC;
	iacseBuff[1] = SE;
	SendReply (iacseBuff, 2);
}

void
AuthOption::NegotiateOption(char command)
{
	char c;
	
	switch (command) {
		case DO:
			char buff[3];
			buff[0] = IAC;
			if (KerbInit()) {
				buff[1] = WONT;
			telnet->TerminalNote("[ WARNING WONT use KERBEROS authentication or encrypt. WARNING ].\r\n");
				}
			else {
				buff[1] = WILL;
				}
			buff[2] = option;
			SendReply (buff, 3);
			telnet->MarkAsRead();
			break;
		case SB:
			if (!telnet->RawGetChar(&c))
				return;
			switch (c) {
				case SEND:
					NegotiateSend();
					break;
				case REPLY:
					NegotiateReply();
					break;
				default:
					break;
				}		
			break;
		default:
			break;
	}
}

void
AuthOption::SendNoKerb4()
{
	static char buff[8];

	buff[0] = IAC;
	buff[1] = SB;
	buff[2] = AUTHENTICATION;
	buff[3] = IS;
	buff[4] = NULL;
	buff[5] = 0;
	buff[6] = IAC;
	buff[7] = SE;
	SendReply (buff, 8);
}

void
AuthOption::NegotiateSend()
{
	char rch;
	register int i;

	for (i = 0; i < MAX_AUTH_TYPES; i++) {
		if (!telnet->RawGetChar(&rch))
			return;
		if (rch == IAC)
			break;
		AuthPairs[i].type = rch;
		if (!telnet->RawGetChar(&AuthPairs[i].mod))
			return;
		if (AuthPairs[i].mod == IAC) {
			telnet->UngetChar();
			if (ReadToIACSE())
				return;
			telnet->MarkAsRead();
			SendNoKerb4 ();
			return;
			}
		}
	if (rch != IAC)
		if (ReadToIACSE())
			return;

	telnet->UngetChar();
	if (ReadToIACSE())
		return;
	telnet->MarkAsRead();
	AuthCount = i;
	AuthIdx = 0;

	NegotiateRetry();
}

#pragma argsused
void
AuthOption::NegotiateRetry(const char *reason)
{
	char buf[256];

	AuthMode = enAuthUnk;

	for (;AuthIdx < AuthCount; AuthIdx++) {
		switch (AuthPairs[AuthIdx].type) {
			case KERBEROS_V4:
				strlcpy(buf, "[ Trying ", sizeof(buf));
				if (AuthPairs[AuthIdx].mod & AUTH_HOW_MUTUAL)
					strlcat(buf, "mutual ", sizeof(buf));
				strlcat(buf, "KERBEROS4 ]\r\n", sizeof(buf));
				break;
			case KERBEROS_V5:
				strlcpy(buf, "[ Trying ", sizeof(buf));
				if (AuthPairs[AuthIdx].mod & AUTH_HOW_MUTUAL)
					strlcat(buf, "mutual ", sizeof(buf));
				strlcat(buf, "KERBEROS5 ]\r\n", sizeof(buf));
				break;
			default:
				sprintf(&buf[strlen(buf)],
					"[ Unsupported type %02x, not used ]\r\n",
						AuthPairs[AuthIdx].type & 0xff);
			}
		telnet->TerminalNote(buf);
		switch (AuthPairs[AuthIdx].type) {
			case KERBEROS_V4:
				NegotiateSendKrb4(AuthPairs[AuthIdx].mod);
				break;
			case KERBEROS_V5:
				NegotiateSendKrb5(AuthPairs[AuthIdx].mod);
				break;
			}
		if (AuthMode != enAuthUnk)
			break;
		}
	if (AuthMode == enAuthUnk) {
		SendNoKerb4();
//		telnet->SendRetryConnection(reason);
		}
}

void
AuthOption::NegotiateSendKrb4(char mod)
{
	char padKBuff[2048];
	char kBuff[2048], *pKey, principal[255], hostname[255];
	unsigned long kBuffLen = sizeof (kBuff);
	Kerr result;

	if (mod != AUTH_HOW_MUTUAL)
		return;

	// Get Kerberos ticket
	principal[0] = '\0';

	if (cd.UseTicket[0] == 0) {
		strlcat(principal, "rcmd.", sizeof(principal));
		strlcpy(hostname, cd.HostName, sizeof(hostname));
		char *p = strchr(hostname, '.');
		if (p)
			*p = 0;
		strlcat(principal, hostname, sizeof(principal));
		strlcat(principal,"@", sizeof(principal));
		strlcat(principal, krb_realmofhost(cd.HostName), sizeof(principal));
		}
	else
		strlcpy(principal, cd.UseTicket, sizeof(principal));
	
	strlcpy(kBuff, "[ Using ticket: ", sizeof(kBuff));
	strlcat(kBuff, principal, sizeof(kBuff));
	strlcat(kBuff, " ]\r\n", sizeof(kBuff));
	telnet->TerminalNote(kBuff);

	result = pKClient->GetTicketForService(principal, kBuff, &kBuffLen, 0, true);
	if (result == KFAILURE) {
		telnet->TerminalNote("[ KERBEROS4 authentication cancelled by user ]\r\n");
		return;
		}

	if (result) {
		char error[255];
		pKClient->ErrorText(result, error); 

		strlcpy(noteBuf,
			"[ Unable to get ticket ]\r\n", 
			sizeof(noteBuf));
		strlcat(noteBuf, "[ Reason: ", sizeof(noteBuf));
		strlcat(noteBuf, (char *)error, sizeof(noteBuf));
		strlcat(noteBuf, " ]\r\n", sizeof(noteBuf));
		telnet->TerminalNote(noteBuf);
		AuthIdx++;
		NegotiateRetry(error);
		return;
		}
	
	// username

	if (cd.UserName[0] == '\0') {
		result = pKClient->GetUserName(cd.UserName);
		if (result) {
			char error[255];
			pKClient->ErrorText(result, error);
			pKClient->Message(error, result);
			SendNoKerb4();
			return;
			}
		}

	SendUserId();
		
	// Check if kBufflen in beginning of kBuff
	if (*((unsigned long *)kBuff) == kBuffLen - sizeof (unsigned long)) {
		pKey = kBuff + sizeof (unsigned long);
		kBuffLen -= sizeof (unsigned long);
		}
	else
		pKey = kBuff;
	
	// Send ticket
	unsigned char tktBuff[7];
	char iacseBuff[2];
	tktBuff[0] = IAC;
	tktBuff[1] = SB;
	tktBuff[2] = AUTHENTICATION;
	tktBuff[3] = IS;
	tktBuff[4] = KERBEROS_V4;
	tktBuff[5] = AUTH_CLIENT_TO_SERVER | AUTH_HOW_MUTUAL;
	tktBuff[6] = KRB4_AUTH;
	SendReply (tktBuff, 7);
	
	unsigned long padKBuffLen;
	PadBinaryBuff(pKey, &kBuffLen, padKBuff, &padKBuffLen); 
	SendReply(padKBuff, padKBuffLen);
	iacseBuff[0] = IAC;
	iacseBuff[1] = SE;
	SendReply(iacseBuff, 2);
	AuthMode = enAuthKrb4;
}

void
AuthOption::NegotiateSendKrb5(char mod)
{
	krb5_data cksum_data;
	char foo[2];
	char buf[256];
	krb5_data auth;
	int ap_opts;

	if (!(bool)*k5c)
		return;

	if ((mod & AUTH_HOW_MASK) == AUTH_HOW_MUTUAL)
		ap_opts = AP_OPTS_MUTUAL_REQUIRED;
	else
		ap_opts = 0;
    
	if (!k5c->AuthInitContext()) {
		telnet->TerminalNote(k5c->GetErrorNote(buf,
			"InitAuthContext"));
		return;
		}

	if (!k5c->AuthSetAddrs(&telnet->mSocket)) {
		telnet->TerminalNote(k5c->GetErrorNote(buf,
			"AuthSetAddrsContext"));
		return;
		}

	k5c->AuthSetKeytype(KEYTYPE_DES);

	foo[0] = KERBEROS_V5;
	foo[1] = mod;

	cksum_data.length = sizeof(foo);
	cksum_data.data   = foo;

	if (!k5c->MakeRequestLogin(telnet->Parent, ap_opts, "host",
		cd.HostName, &cksum_data, &auth)) {
		telnet->TerminalNote(k5c->GetErrorNote(buf,
			"MakeRequestLogin"));
		return;
		}

	if (cd.UserName[0] == 0) {
		strlcpy(cd.UserName, k5c->GetUser(), sizeof(cd.UserName));
#if 0
		krb5_principal principal;
		char *s;

		if (k5c->CCGetPrincipal(&principal)) {
			if (k5c->UnparseNameUser(principal, &s)) {
				strlcpy(cd.UserName, s, sizeof(cd.UserName));
				free(s);
				}
			k5c->FreePrincipal(principal);
			}
#endif

		}

	SendUserId();

	Data(mod, KRB5_AUTH, auth.data, auth.length);

	telnet->TerminalNote("[ Sent Kerberos V5 credentials to server ]\r\n");
	AuthMode = enAuthKrb5;
}

void
AuthOption::SendChallengeKrb4()
{
	int i;
	unsigned char reply1[] = {
		IAC, SB, AUTHENTICATION, IS, KERBEROS_V4,
		AUTH_CLIENT_TO_SERVER | AUTH_HOW_MUTUAL, KRB4_CHALLENGE};
	char reply2[] = { IAC, SE };
	char msg[16];
	unsigned long len, padlen;
	des_cblock session;
	des_key_schedule schedule;

	srand(time(0));
	for (i = 0; i < 8; i++)
		telnet->enckey[i] = (unsigned char)(rand() & 0xff);
	des_set_odd_parity((des_cblock *)telnet->enckey);
	pKClient->GetSessionKey((KClientKey*)&session);
	des_set_key(&session, schedule);
	des_ecb_encrypt(&telnet->enckey, (des_cblock *)&challenge, schedule,
		DES_DECRYPT);

	len = 8;
	PadBinaryBuff(challenge, &len, msg, &padlen);

	SendReply(reply1, sizeof(reply1));
	SendReply(msg, padlen);
	SendReply(reply2, sizeof(reply2));

	des_ecb_encrypt((des_cblock *)&challenge, (des_cblock *)&challenge,
		schedule, DES_DECRYPT);

#pragma warn -sig
	for (i = 7; i >= 0; i--)
		if (++(challenge[i]) != 0)
			break;
#pragma warn .sig

	des_ecb_encrypt((des_cblock *)&challenge, (des_cblock *)&challenge,
		schedule, DES_ENCRYPT);

	memset(&session, 0, sizeof(des_cblock));
	memset(&msg, 0, sizeof(msg));
	memset(&schedule, 0, sizeof(des_key_schedule));
}


void
AuthOption::NegotiateReply()
{
	char pair[2];

	if (!(telnet->RawGetChar(pair)) ||
	    !(telnet->RawGetChar(pair + 1))) // Get auth-pair
		return;
	
	switch (pair[0]) {
		case KERBEROS_V4:
			NegotiateReplyKrb4(pair[1]);
			break;
		case KERBEROS_V5:
			NegotiateReplyKrb5(pair[1]);
			break;
		default:
			if (ReadToIACSE())
				return;
			telnet->MarkAsRead();
			SendNoKerb4();
			return;
		}
}

void
AuthOption::NegotiateReplyKrb4(char mod)
{
	char reply;
	unsigned char reason[80];
	short i;

	if (!telnet->RawGetChar(&reply))
		return;
	
	switch (reply) {
		case KRB4_REJECT:
			if (GetToIAC(reason, sizeof(reason)) < 0)
				return;

			strlcpy(noteBuf,
				"[ Kerberos V4 didn't accept you ]\r\n",
				sizeof(noteBuf));
			strlcat(noteBuf,
				"[ Reason: ",
				sizeof(noteBuf));
			strlcat(noteBuf,
				(char *)reason,
				sizeof(noteBuf));
			strlcat(noteBuf, " ]\r\n", sizeof(noteBuf));
			telnet->TerminalNote(noteBuf);
			telnet->UngetChar();
			telnet->SendRetryConnection((char *)reason);
			break;	
		case KRB4_ACCEPT:
			if (mod & AUTH_HOW_MUTUAL) {
				telnet->TerminalNote("[ Kerberos V4 accepts you ]\r\n");
				SendChallengeKrb4();
				}
			break;
		case KRB4_RESPONSE:
			des_cblock response;
			// Get encrypted response
			for (i = 0; i < 8; i++)
				if (!telnet->RawGetChar((char *)response + i))
					return;
			for (i = 0; i < 8; i++)
				if (response[i] != challenge[i])
					break; //return;
			telnet->TerminalNote("[ Kerberos V4 challenge successful ]\r\n");
			telnet->mAuthenticated = true;
			NegotiateEncryption();
			break;
		default:
			break;
		}

	if (ReadToIACSE())
		return;
	else
		telnet->MarkAsRead();
}

void
AuthOption::NegotiateReplyKrb5(char mod)
{
	char reply;
	unsigned char arg[512];
	char buf[256];
	short arglen;
	krb5_keyblock *keyblock;

	if (!telnet->RawGetChar(&reply))
		return;
	
	if ((arglen = GetToIAC(arg, sizeof(arg))) < 0)
		return;

	switch (reply) {
		case KRB5_REJECT:
			strlcpy(noteBuf,
				"[ Kerberos V5 didn't accept you ]\r\n",
				sizeof(noteBuf));
			strlcat(noteBuf,
				"[ Reason: ",
				sizeof(noteBuf));
			strlcat(noteBuf,
				(char *)arg,
				sizeof(noteBuf));
			strlcat(noteBuf, " ]\r\n", sizeof(noteBuf));
			telnet->TerminalNote(noteBuf);
			telnet->UngetChar();
			AuthIdx++;
			NegotiateRetry((char *)arg);
			break;	
		case KRB5_RESPONSE:
			telnet->TerminalNote("[ Kerberos V5 Response received ]\r\n");
			if ((mod & AUTH_HOW_MASK) == AUTH_HOW_MUTUAL) {
				/* the rest of the reply should contain */
				/* a krb_ap_rep */
				krb5_ap_rep_enc_part *k5_reply;
	    
				if (!k5c->ReadReply(K5Data(arg, arglen),
					&k5_reply)) {
					telnet->TerminalNote(k5c->GetErrorNote(
						buf, "Authentication"));
					return;
					}
				k5c->FreeReply(k5_reply);
				telnet->mAuthenticated = true;
				NegotiateEncryption();
				}
			break;
		case KRB5_ACCEPT:
			if (mod & AUTH_HOW_MUTUAL) {
				strlcpy(noteBuf,
					"[ Kerberos V5 accepts you",
					sizeof(noteBuf));
				if (arglen) {
					strlcat(noteBuf, " as '",
						sizeof(noteBuf));
					strlcat(noteBuf, (char *)arg,
						sizeof(noteBuf));
					strlcat(noteBuf, "'",
						sizeof(noteBuf));
					}
				strlcat(noteBuf, " ]\r\n",
					sizeof(noteBuf));
				telnet->TerminalNote(noteBuf);
				
				if (!k5c->AuthGetLocalSubKey(&keyblock))
					if (!k5c->AuthGetLocalKey(&keyblock)) {
						telnet->TerminalNote(
							k5c->GetErrorNote(
								buf,
								"GetKey"));
						return;
						}

				MoveMemory(telnet->enckey,
					keyblock->keyvalue.data,
					8);
				k5c->FreeKeyblock(keyblock);
				}
			break;
		default:
			sprintf(noteBuf, "[ Unknown reply: %d ]\r\n", reply);
			telnet->TerminalNote(noteBuf);
			break;
		}

	telnet->MarkAsRead();
}

void
AuthOption::NegotiateEncryption()
{
	unsigned char mBuf[] = {IAC, DO, ENCRYPT_OPT};
	SendReply(mBuf, 3);
	mBuf[1] = WILL;
	SendReply(mBuf, 3);
}

bool
AuthOption::InEffect()
{
	return false;
}
