/*
 * Copyright (c) 2000 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <krb5.h>

class _KTCOMMONCLASS K5Data : public krb5_data
{
      public:
	K5Data(void *arg, int arglen) {
		length = (size_t)arglen;
		data = arg;
		};
};

class _KTCOMMONCLASS Kerberos5Context
{
      public:
	Kerberos5Context(SetupParam *setupData);
	~Kerberos5Context();
	char * GetErrorNote(char *buf, const char *note);
	const char * GetErrorText();
	bool OpenLog(const char *program);
	bool Log(const char *fmt, ...);

	bool GetNoAddressesFlag() const;

	bool AuthInitContext();
	bool AuthSetAddrs(SOCKET *Socket);
	void AuthSetKeytype(krb5_keytype keytype);
	bool AuthGetLocalSubKey(krb5_keyblock **keyblock);
	bool AuthGetLocalKey(krb5_keyblock **keyblock);
	void FreeKeyblock(krb5_keyblock *keyblock);

	bool GetTicketForRealms(TWindow *window, krb5_realm *realms);
	bool MakeRequestLogin(TWindow *window, krb5_principal princ);
	bool MakeRequestLogin(TWindow *window,
		krb5_flags ap_req_options,
		const char *service, const char *hostname,
		krb5_data *indata, krb5_data *outbuf);
	bool MakeRequest(krb5_flags ap_req_options,
		const char *service, const char *hostname,
		krb5_data *indata, krb5_data *outbuf);
	bool ReadReply(krb5_data inbuf, krb5_ap_rep_enc_part **k5_reply);
	void FreeReply(krb5_ap_rep_enc_part *reply);

	void GetInitCredsOptInit();
	void GetInitCredsOptSetAddressList(krb5_addresses *addrlist);
	bool GetInitCredsPassword(krb5_creds *cred,
		krb5_principal principal, const char *password);

	bool CCInit(krb5_principal primary_principal);
	bool CCStoreCred(krb5_creds *cred);
	bool CCGetPrincipal(krb5_principal *principal);
	bool CCGetPrincipalRealm(const char *realm,
		krb5_principal *principal);
	bool CCStartSeqGet(krb5_cc_cursor *cursor);
	bool CCNextCred(krb5_creds *cred, krb5_cc_cursor *cursor);
	void CCEndSeqGet(krb5_cc_cursor *cursor);
	bool CCDestroy();

	bool GetCredentials(krb5_creds *incred, krb5_creds **outcred);
	bool FreeCredsContents(krb5_creds *cred);

	bool Conv4to5Principal(KCPBuf &kb, krb5_principal *principal);
	bool UnparseName(krb5_const_principal principal, char **str);
	bool UnparseNameShort(krb5_const_principal principal, char **str);
	bool UnparseNameUser(krb5_const_principal principal, char **str);
	bool UnparseNameUserInst(krb5_const_principal principal, char **str);

	void FreePrincipal(krb5_principal principal);

	bool HaveLoginTicket(krb5_realm *realms);
	bool Login(TWindow *parent, krb5_realm default_realm);

	operator bool() const;
	operator krb5_context() const;

	krb5_error_code ret;

	const char *GetUser();
	KCPBuf GetKCPBuf();

      protected:
	krb5_context context;
	krb5_ccache  ccache;
	krb5_auth_context auth_context;
	bool Krb5OK;
	krb5_get_init_creds_opt opt;
	krb5_log_facility *logfac;
	KCPBuf kd;
	bool NoAddresses;

	SetupParam *SetupData;
};


inline Kerberos5Context::operator bool() const
{
	return Krb5OK;
}

inline Kerberos5Context::operator krb5_context() const
{
	return context;
}

inline const char *
Kerberos5Context::GetUser()
{
	return kd.user;
}

inline KCPBuf
Kerberos5Context::GetKCPBuf()
{
	return kd;
}

inline bool
Kerberos5Context::GetNoAddressesFlag() const
{
	return NoAddresses;
}
