/*
 * Copyright (c) 2000 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define STRICT
#define _WXRTL_DLL
#include "wxrtl.h"
#pragma hdrstop

#include <roken.h>
#include <owl\dialog.h>
#include <owl\radiobut.h>
#include <owl\checkbox.h>
#include <owl\edit.h>
#include <owl\combobox.h>
#include <owl\opensave.h>
#include <string.h>
#include "edititem.h"

#include "wxrtl.rh"

//----------------------------------------------------------------

class TreeEditItem : public EditItem
{
      public:
	TreeEditItem(TWindow *parent, TListWindow *lw, int Index,
		EPTNode *_SelectedTree, bool EditLabel = false);
	virtual void CloseWindow(int code);
	
      private:
	EPTNode *SelectedTree;
};

TreeEditItem::TreeEditItem(TWindow *parent, TListWindow *lw, int index, 
	EPTNode *_SelectedTree, bool editLabel)
	: EditItem(parent, lw, index, editLabel)
{
	SelectedTree = _SelectedTree;
}

void
TreeEditItem::CloseWindow(int code)
{
	if (code == IDOK) {
		if (SelectedTree == 0) {
			MessageBeep(-1);
			return;
			}

		TransferData(tdGetData);

		if (EditLabel) {
			if (strlen(trfbuf.label) == 0) {
				MessageBeep(-1);
				return;
				}
			LW->SetItemText(Index, TListWindItem(trfbuf.label));
			}

		LW->SetItemText(Index, TListWindItem(trfbuf.value, 1));
		TListWindItem item;
		item.SetIndex(Index);
		item.mask |= LVIF_PARAM;		
		LW->GetItem(item);
		EPTNode *pp = (EPTNode *)item.GetItemData();

		if (!pp) {
			SelectedTree->epth->AddNode(EPTNode::vkValue,
				SelectedTree, trfbuf.label, trfbuf.value);
			}
		else {
			delete[] pp->u.value;
			pp->u.value = strnewdup(trfbuf.value);
			}
		}
	// Don't go through EditItem's CloseWindow!
	TDialog::CloseWindow(code);
}

//----------------------------------------------------------------

EPTHelper::EPTHelper()
{
	flags = 0;
}

EPTHelper::~EPTHelper()
{
}

void
EPTHelper::ExportToFile(TWindow * /*parent*/, EPTNode * /*tree*/)
{
	MessageBeep(-1);
}

void
EPTHelper::ImportFromFile(TWindow * /*parent*/, EPTNode * /*tree*/)
{
	MessageBeep(-1);
}

EPTNode *
EPTHelper::FindNode(EPTNode::tc type, EPTNode *list, const char *name)
{
	for (;list; list = list->next) {
		if (list->type != type)
			continue;
		if (strcasecmp(name, list->name) == 0)
			return list;
		}
	return NULL;
}

EPTNode *
EPTHelper::AddNode(EPTNode::tc type, EPTNode *root, const char *name,
	const char *value)
{
	EPTNode *v = new EPTNode;
	v->type = type;
	v->next = root->u.node;
	v->name = strnewdup(name);
	if (type == EPTNode::vkValue)
		v->u.value = strnewdup(value);
	else
		v->u.node = 0;
	v->level = root->level + 1;
	v->epth = root->epth;
	root->u.node = v;
	return v;
}

//----------------------------------------------------------------

DEFINE_RESPONSE_TABLE1(EditPropertyTree, SetupPage)
	EV_WM_CLOSE,
	EV_WM_CONTEXTMENU,
	EV_TVN_SELCHANGED(IDC_EPTTREE, SelChanged),
	EV_TVN_BEGINLABELEDIT(IDC_EPTTREE, BeginTreeLabelEdit),
	EV_TVN_ENDLABELEDIT(IDC_EPTTREE, EndTreeLabelEdit),

	EV_LVN_ENDLABELEDIT(IDC_EPTLIST, EndListLabelEdit),
	EV_LVN_ITEMCHANGED(IDC_EPTLIST, SelChanged),

	EV_COMMAND(IDCANCEL, CmCancel),
	EV_COMMAND(CM_PLDELETE, CmDeleteItem),
	EV_COMMAND(CM_PLRENAME, CmRenameItem),
	EV_COMMAND(CM_PLVALUE, CmValueItem),
	EV_COMMAND(CM_PLADD, CmAddItem),
	EV_COMMAND(IDC_PLIMPORT, CmImport),
	EV_COMMAND(IDC_PLEXPORT, CmExport),

END_RESPONSE_TABLE;

EditPropertyTree::EditPropertyTree(PropertySetup *ns, TPropertySheet *ps,
	SetupPart *sp)
	: SetupPage(ns, ps, sp)
{
	tw = new TTreeWindow(this, IDC_EPTTREE);
	lw = new TListWindow(this, IDC_EPTLIST);
	tree = 0;
	LabelEditMode = leNone;
	MenuWindow = msNone;
	WriteEnabled = true;
	SelectedItem = -1;
	SelectedTree = 0;
}

EditPropertyTree::~EditPropertyTree()
{
	FreeTree(&tree);
}

void
EditPropertyTree::AddSubTree(Registry *root, const char *subkey,
	EPTHelper *epth)
{
	EPTNode *v = new EPTNode;
	v->type = EPTNode::vkRootNode;
	v->next = tree;
	v->name = strnewdup(subkey);
	v->u.node = 0;
	v->reg = new Registry(root, subkey);
	v->epth = epth;
	v->level = 0;
	tree = v;
}

void
EditPropertyTree::SetupWindow()
{
	char key[1024];
	EPTNode *w;

	TDialog::SetupWindow();
	
	TPopupMenu *pm = new TPopupMenu(TMenu(*ThisModule, IDM_PRPLIST));
	tw->DeleteAllItems();

	lw->InsertColumn(0, TListWindColumn("Parameter",
		lw->GetStringWidth("ParameterName") + 30));
	lw->InsertColumn(1, TListWindColumn("Value",
		lw->GetStringWidth("ThisIsAStringUsedForInformation") + 30));

	if (pm->GetMenuString(CM_PLRENAME, key, sizeof(key),
		MF_BYCOMMAND)) {
		strcat(key, " ");
		strcat(key, "Parameter name");
		pm->ModifyMenu(CM_PLRENAME, MF_BYCOMMAND, CM_PLRENAME, key);
		}

	if (pm->GetMenuString(CM_PLVALUE, key, sizeof(key),
		MF_BYCOMMAND)) {
		strcat(key, " ");
		strcat(key, "Value");
		pm->ModifyMenu(CM_PLVALUE, MF_BYCOMMAND, CM_PLVALUE, key);
		}

	
	for (w = tree; w; w = w->next) {
		TTreeItem item(w->name);
		item.SetItemData((uint32)w);
		AddTree(tw->GetRoot().AddChild(item), *w->reg, w);
		}

	AssignContextMenu(pm);
}

int
EditPropertyTree::Apply(TNotify far &)
{
	EPTNode *w;

	if (LabelEditCheck())
		return PSNRET_INVALID_NOCHANGEPAGE;

	for (w = tree; w; w = w->next) {		
		if (!SaveInRegistry(*w->reg, w->u.node))
			return PSNRET_INVALID_NOCHANGEPAGE;
		}

	return PSNRET_NOERROR;
}

void
EditPropertyTree::CmCancel()
{
	EvClose();
}

void
EditPropertyTree::EvClose()
{
	if (LabelEditCheck(true))
		return;
	
	TDialog::EvClose();
}

bool
EditPropertyTree::LabelEditCheck(bool Cancel)
{
	switch (LabelEditMode) {
		case leTree:
			tw->SendMessage(TVM_ENDEDITLABELNOW, Cancel);
			return true;
		case leList:
		default:
			break;
		}
	return false;
}

void
EditPropertyTree::EvContextMenu(HWND child, int x, int y)
{
	TPopupMenu *pm;
	uint en;

	if (child == *lw)
		MenuWindow = msList;
	else if (child == *tw)
		MenuWindow = msTree;
	else
		MenuWindow = msNone;

	if ((MenuWindow != msNone) && (pm = GetContextMenu()) != 0) {
		pm->EnableMenuItem(CM_PLADD, MF_ENABLED);
		pm->EnableMenuItem(CM_PLDELETE, MF_ENABLED);
		pm->EnableMenuItem(IDC_PLIMPORT, MF_GRAYED);
		pm->EnableMenuItem(IDC_PLEXPORT, MF_GRAYED);

		if (!WriteEnabled) {
			en = MF_GRAYED;
			pm->EnableMenuItem(IDC_PLIMPORT, MF_GRAYED);
			pm->EnableMenuItem(CM_PLADD, MF_GRAYED);
			}
		else
			en = (SelectedItem != -1)? MF_ENABLED
						 : MF_GRAYED;
		pm->EnableMenuItem(CM_PLRENAME, en);
		pm->EnableMenuItem(CM_PLVALUE, en);

		if (MenuWindow == msTree) {
			if (!SelectedTree) {
				pm->EnableMenuItem(CM_PLDELETE, MF_GRAYED);
				pm->EnableMenuItem(CM_PLADD, MF_GRAYED);
				}
			else switch (SelectedTree->level) {
				case 0:		// Root
					pm->EnableMenuItem(CM_PLDELETE,
						MF_GRAYED);
					if ((SelectedTree->epth->flags &
						(EPTF_TWO_LEVEL|EPTF_MANY_LEVELS)) == 0)
						pm->EnableMenuItem(CM_PLADD,
							MF_GRAYED);
					pm->EnableMenuItem(IDC_PLIMPORT,
						MF_ENABLED);
					pm->EnableMenuItem(IDC_PLEXPORT,
						MF_ENABLED);
					break;
				case 1:		// First level
				default:	// Deep levels
					if ((SelectedTree->epth->flags &
						EPTF_MANY_LEVELS) == 0)
						pm->EnableMenuItem(CM_PLADD,
							MF_GRAYED);
					break;
				}
			}

		pm->TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON,
			x, y, 0, *this, 0);
		return;
		}
	DefaultProcessing();
}

void
EditPropertyTree::CmRenameItem()
{
	if (SelectedItem != -1)
		lw->EditLabel(SelectedItem);
}

void
EditPropertyTree::CmValueItem()
{
	if (SelectedItem == -1) {
		MessageBeep(-1);
		return;
		}

	TreeEditItem(this, lw, SelectedItem, SelectedTree).Execute();
}

//****************************************************************
//
//	Events handling
//
//****************************************************************

void
EditPropertyTree::SelChanged(TTwNotify &n)
{
	char buf[128];
	TTreeItem item;

	lw->DeleteAllItems();

	item.SetHTreeItem(n.itemNew.hItem);
	item.SetText(buf, sizeof(buf));
	item.mask |= TVIF_CHILDREN|TVIF_PARAM;

	if (!tw->SendMessage(TVM_GETITEM, 0, TParam2(&item))) {
		SelectedTree = 0;
		return;		// Can't expand!
		}

	EPTNode *v = (EPTNode *)item.GetItemData();

	SelectedTree = v;
	SelectedItem = -1;
	LoadListWindow(v->u.node);
}

void
EditPropertyTree::SelChanged(TLwNotify &n)
{
	if (n.uNewState & LVIS_SELECTED)
		SelectedItem = n.iItem;
	else if (n.iItem == SelectedItem)
		SelectedItem = -1;
}

bool
EditPropertyTree::BeginTreeLabelEdit(TTwDispInfoNotify &n )
{
	EPTNode *e = (EPTNode *)(n.item.lParam);
	if (!e)
		return true;

	if (e->level == 0)
		return true;

	LabelEditMode = leTree;
	return false;
}

void
EditPropertyTree::EndTreeLabelEdit(TTwDispInfoNotify &n)
{
	LabelEditMode = leNone;

	if (n.item.pszText == 0)
		return;			// Cancelled

	if (strlen(n.item.pszText) == 0) {
		MessageBeep(-1);
		return;
		}

	TTreeNode node(*tw, n.item.hItem);
	char OldName[128];
	TTreeItem titem(n.item);
	
	titem.SetText(OldName, sizeof(OldName));
	titem.mask |= TVIF_PARAM;
	node.GetItem(&titem);

	TTreeNode w = GetTreeNode(n.item.pszText, node.GetParent());
	
	if (w) {
		MessageBeep(-1);
		HWND ec = node.EditLabel();
		LabelEditMode = leTree;
		::SendMessage(ec, WM_SETTEXT, 0, TParam2(n.item.pszText));
		::SendMessage(ec, EM_SETSEL, 0, strlen(n.item.pszText));
		return;
		}

	EPTNode *pp = (EPTNode *)titem.GetItemData();
	if (pp) {
		delete[] pp->name;
		pp->name = strnewdup(n.item.pszText);
		titem.SetText(n.item.pszText);
		node.SetItem(&titem);
		}
	else
		MessageBeep(-1);
}

void
EditPropertyTree::EndListLabelEdit(TLwDispInfoNotify &n)
{
	LabelEditMode = leNone;

	if ((n.item.iItem == -1) || (n.item.pszText == 0))
		return;			// Cancelled

	if (strlen(n.item.pszText) == 0) {
		MessageBeep(-1);
		return;
		}

	TLwFindInfo fi;
	int idx;

	fi.SetString(n.item.pszText);
	
	for (idx = -1; (idx = lw->FindItem(idx, &fi)) != -1;) {
		if (idx != n.item.iItem)
			break;
		}

	if (idx == -1) {
		TListWindItem item;
		item.SetIndex(n.item.iItem);
		item.mask |= LVIF_PARAM;
		lw->GetItem(item);

		EPTNode *pp = (EPTNode *)item.GetItemData();

		if (pp) {
			delete[] pp->name;
			pp->name = strnewdup(n.item.pszText);
			item.SetText(n.item.pszText);
			lw->SetItem(item);
			}
		else
			MessageBeep(-1);
		}
	else if (idx != n.item.iItem) {
		MessageBeep(-1);
		lw->SetFocus();
		HWND ec = lw->EditLabel(n.item.iItem);
		LabelEditMode = leList;
		::SendMessage(ec, WM_SETTEXT, 0, TParam2(n.item.pszText));
		::SendMessage(ec, EM_SETSEL, 0, strlen(n.item.pszText));
		return;
		}
}

void
EditPropertyTree::CmAddItem()
{
	char key[256];

	switch (MenuWindow) {
		case msList:
			SelectedItem = 
				lw->InsertItem(TListWindItem((char *)ResString(this, IDSR_NEW)));
			lw->SetItem(TListWindItem((char *)ResString(this, IDSR_NEW), 1));
			TreeEditItem(this, lw, SelectedItem, SelectedTree,
				true).Execute();	
			break;
		case msTree:
			TTreeNode node(*tw);
			if (SelectedTree == 0) {
				MessageBeep(-1);
				return;
				}
			node = tw->GetSelection();
			EPTNode *v = SelectedTree->epth->AddNode(
				EPTNode::vkNode, SelectedTree, "New", NULL);
			TTreeItem TreeItem(key);
			TreeItem.SetItemData((uint32)v);

			TTreeNode NewNode(node.AddChild(TreeItem));

			if (node != TVI_ROOT) {
				TTreeItem item;
				item.mask |= TVIF_CHILDREN;
				node.GetItem(&item);
				item.cChildren++;
				node.SetItem(&item);
				}

			NewNode.SelectItem(TTreeNode::Caret);
			HWND ec = NewNode.EditLabel();
			LabelEditMode = leTree;
			::SendMessage(ec, WM_SETTEXT, 0,
				TParam2(v->name));
			::SendMessage(ec, EM_SETSEL, 0,
				strlen(v->name));
			break;
		}
		
}

void
EditPropertyTree::CmExport()
{
	if ((MenuWindow != msTree) || (!SelectedTree)) {
		MessageBeep(-1);
		return;
		}
	
	if ((SelectedTree->type != EPTNode::vkRootNode) ||
	    (SelectedTree->epth == 0)) {
		MessageBeep(-1);
		return;
		}
	SelectedTree->epth->ExportToFile(this, SelectedTree);
}

void
EditPropertyTree::CmImport()
{
	if ((MenuWindow != msTree) || (!SelectedTree)) {
		MessageBeep(-1);
		return;
		}
	
	if ((SelectedTree->type != EPTNode::vkRootNode) ||
	    (SelectedTree->epth == 0)) {
		MessageBeep(-1);
		return;
		}
	SelectedTree->epth->ImportFromFile(this, SelectedTree);

	TTreeNode node(*tw);
	node = tw->GetSelection();

	TTreeNode child(*tw);

	do {
		child = node.GetChild();
		if (child)
			child.Delete();
	} while (child);

	TTreeItem item;
	item.mask |= TVIF_CHILDREN;
	item.cChildren = 0;
	node.SetItem(&item);
	RebuildTree(node, SelectedTree->u.node);

	lw->DeleteAllItems();
	SelectedItem = -1;
	LoadListWindow(SelectedTree->u.node);
}

void
EditPropertyTree::RebuildTree(TTreeNode node, EPTNode *v)
{
	TTreeItem item;
	item.mask |= TVIF_CHILDREN;
	item.cChildren = 0;

	for (; v; v = v->next) {
		if (v->type == EPTNode::vkNode) {
			TTreeItem TreeItem(v->name);
			TreeItem.SetItemData((uint32)v);
			RebuildTree(node.AddChild(TreeItem), v->u.node);
			item.cChildren++;
			}
		}
	node.SetItem(&item);
}

void
EditPropertyTree::CmDeleteItem()
{
	char label[64], buf[256];
		
	switch (MenuWindow) {
		case msList: {
			if (SelectedItem == -1) {
				MessageBeep(-1);
				return;
				}

			TListWindItem item(label, 0, sizeof(label));
			item.SetIndex(SelectedItem);
			item.mask |= TVIF_PARAM;
			lw->GetItem(item);

			sprintf(buf, ResString(this, IDSR_LABDEL), "Parameter", label);
			if (MessageBox(buf, ResString(this, IDSR_CONFDEL),
				MB_YESNO|MB_ICONQUESTION) == IDYES) {
				EPTNode *pp = (EPTNode *)item.GetItemData();
				if (pp)
					pp->type = EPTNode::vkDeletedValue;
				lw->DeleteAnItem(SelectedItem);
				SelectedItem = -1;
				}
			} break;
		case msTree:
			if (!SelectedTree) {
				MessageBeep(-1);
				return;
				}
			snprintf(buf, sizeof(buf),
				ResString(this, IDSR_LABDEL),
				"Group", SelectedTree->name);
			if (MessageBox(buf, ResString(this, IDSR_CONFDEL),
				MB_YESNO|MB_ICONQUESTION) == IDYES) {
				SelectedTree->type = EPTNode::vkDeletedNode;
				SelectedTree = 0;

				TTreeNode node(tw->GetSelection());
				node.Delete();
				}
			break;
		}
}

//****************************************************************
//
//	Data handling
//
//****************************************************************

void
EditPropertyTree::AddTree(TTreeNode node, Registry &r, EPTNode *parent)
{
	register int i;
	char key[128];
	char value[512];

	try {
		r.Open();
		}
	catch (RegistryFail) {
		return;
		}
	try {
		for (i = 0; ;i++) {
			r.EnumKey(i, key, sizeof(key));
			EPTNode *v = parent->epth->AddNode(EPTNode::vkNode,
				parent, key, NULL);
			TTreeItem TreeItem(key);
			TreeItem.SetItemData((uint32)v);
			AddTree(node.AddChild(TreeItem), Registry(&r, key),
				v);
			}
		}
	catch (RegistryFail) {
		}

	if (node != TVI_ROOT) {
		TTreeItem item;
		item.cChildren = i;
		item.mask |= TVIF_CHILDREN;
		node.SetItem(&item);
		}

	try {
		for (i = 0; ;i++) {
			r.EnumString(i, key, sizeof(key),
				value, sizeof(value));
			parent->epth->AddNode(EPTNode::vkValue,
				parent, key, value);
			}
		}
	catch (RegistryFail) {
		}

	r.Close();
}

void
EditPropertyTree::FreeTree(EPTNode **pp)
{
	while (*pp) {
		EPTNode *p = *pp;
		switch (p->type) {
			case EPTNode::vkRootNode:
				delete p->reg;
				if (p->epth)
					delete p->epth;
			case EPTNode::vkNode:
			case EPTNode::vkDeletedNode:
				FreeTree(&p->u.node);
				break;
			case EPTNode::vkValue:
			case EPTNode::vkDeletedValue:
				if (p->u.value)
					delete[] p->u.value;
				break;
			}
		*pp = p->next;
		delete[] p->name;
		delete p;
		}
}
void
EditPropertyTree::LoadListWindow(EPTNode *v)
{
	TListWindItem ListItem;

	for (; v; v = v->next) {
		if (v->type != EPTNode::vkValue)
			continue;
		ListItem.mask = 0;
		ListItem.SetItemData((uint32)v);
		ListItem.SetSubItem(0);
		lw->InsertItem(ListItem);
		ListItem.SetText(v->name);
		lw->SetItem(ListItem);

		ListItem.mask &= ~LVIF_PARAM;
		ListItem.lParam = 0;
		ListItem.SetSubItem(1);
		ListItem.SetText(v->u.value);
		lw->SetItem(ListItem);
		}
}

TTreeNode
EditPropertyTree::GetTreeNode(const char *name, HTREEITEM ParentItem)
{
	TTreeNode node(*tw, ParentItem);
	TTreeNode w(*tw);
	char buf[128];
	
	w = node.GetChild();
	while (w) {
		TTreeItem item;
			
		item.SetText(buf, sizeof(buf));
		w.GetItem(&item);
		if (strcmp(buf, name) == 0)
			break;
		w = w.GetNextSibling();
		}
	return w;
}

bool
EditPropertyTree::SaveInRegistry(Registry &r, EPTNode *p)
{
	char l[64], v[1024];
	register int i;
	EPTNode *w;

	// First remove all simple values that are removed

	try {
		r.Open();
		for (i = 0; ; i++) {
			if (r.EnumString(i, l, sizeof(l),
				v, sizeof(v)) == REG_SZ) {
				
				for (w = p; w; w = w->next) {
					if (w->type != EPTNode::vkValue)
						continue;
					if (strcasecmp(w->name, l) == 0)
						break;
					}
				if (!w) {
					r.DeleteValue(l);
					i = -1;	// Restart loop
					}
				}
			}
		}
	catch (RegistryFail) {}

	// Then remove all subtrees that are removed

	try {
		for (i = 0; ; i++) {
			r.EnumKey(i, l, sizeof(l));
			for (w = p; w; w = w->next) {
				if (w->type != EPTNode::vkNode)
					continue;
				if (strcasecmp(w->name, l) == 0)
					break;
				}
			if (!w) {
				r.DeleteKey(l);
				i = -1;	// Restart loop
				}
			}
		}
	catch (RegistryFail) {}

	r.Close();

	// Add/Update subtrees and values

	try {
		r.Open(true);

		for (w = p; w ; w = w->next) {
			switch (w->type) {
				case EPTNode::vkNode:
					if (!SaveInRegistry(
						Registry(&r, w->name),
						w->u.node)) {
						r.Close();
						return false;
						}
					break;
				case EPTNode::vkValue:
					r.SetString(w->name,
						w->u.value);
					break;
				}
			}
		}

	catch (RegistryFail) {
		sprintf(v, ResString(this, IDSR_REGERR), l);

		MessageBox(v, ResString(this, IDSR_DATERR),
			MB_OK|MB_ICONEXCLAMATION);
		r.Close();
		return false;
		}
	r.Close();
	return true;
}
